/*
 * Convolution and image filtering using gaussian blur.
 *
 * Grisha Sitkarev, 2011 (c)
 *
 * Use this software under the terms of GNU GPLv2 or later version.
 *
 */
#include <alloca.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <math.h>

#include <gdk-pixbuf/gdk-pixbuf.h>

/* Smooth gaussian filter with sigma = 1. */
float
gauss(int x)
{
	double a, b;

	a = 1.0 / sqrt(2.0*M_PI);
	b = exp(-(x*x) * 0.5);

	return a * b;
}

/* Slow and naive textbook 2D convolution. */
void
filter_image_s(float *img, float *out, int w, int h, float (*flt)(int), int r)
{
	int x, y;
	int i, j;

	for (y = r; y < h-r; y++) {
		for (x = r; x < w-r; x++) {
			out[y*w + x] = 0.0;
			for (i = -r; i <= r; i++) {
				for (j = -r; j <= r; j++) {
					out[y*w + x] += flt(i)*flt(j)*img[(y-i)*w + (x-j)];
				}
			}
		}
	}

}

/* Faster convolution in 2D for separable filters. */
void
filter_image(float *img, float *out, int w, int h, float (*flt)(int), int r)
{
	int x, y;
	int i;
	float *s;

	/* Temporary space for S[k]. */
	s = alloca(sizeof(float) * w);

	memset(out, 0, (w * h) * sizeof(float));

	for (y = r; y < h-r; y++) {

		memset(s, 0, sizeof(float) * w);

		for (x = 0; x < w; x++) {
			for (i = -r; i <= r; i++)
				s[x] += flt(i)*img[(y-i)*w + x];
		}

		for (x = r; x < w-r; x++) {
			for (i = -r; i <= r; i++)
				out[y*w + x] += flt(i)*s[x-i];
		}
	}
}

int
main(int argc, char *argv[])
{
	GdkPixbuf *pbuf;
	GdkColorspace cs;
	GError *gerror;
	unsigned char *pix, *pp;
	int w, h, nr_pixels;
	float *r, *g, *b;
	float *rout, *gout, *bout;
	int ok, i, j, rowpix;

	if (argc < 2)
		exit(1);

	g_type_init();

	/* Load image from file. */
	gerror = NULL;
	pbuf = gdk_pixbuf_new_from_file(argv[1], &gerror);

	if (pbuf == NULL) {
		printf("can't open file %s: %s", argv[1], gerror->message);
		exit(1);
	}
	
	/* Obtain height, width and colorspace (currently only RGB). */
	h = gdk_pixbuf_get_height(pbuf);
	w = gdk_pixbuf_get_width(pbuf);
	cs = gdk_pixbuf_get_colorspace(pbuf);

	/* Alpha-channel forms 4 bytes in RGBA form. */
	if (gdk_pixbuf_get_has_alpha(pbuf))
		rowpix = 4;
	else
		rowpix = 3;

	nr_pixels = w * h;

	/* Obtain pointer to pixels in the buffer. */
	pix = gdk_pixbuf_get_pixels(pbuf);

	printf("%dx%d pixels, rowstride=%d\n", w, h, gdk_pixbuf_get_rowstride(pbuf));

	r = malloc(sizeof(float) * nr_pixels * 3);
	g = r + nr_pixels;
	b = g + nr_pixels;

	rout = malloc(sizeof(float) * nr_pixels * 3);
	gout = rout + nr_pixels;
	bout = gout + nr_pixels;

	pp = pix;

	/* Copy R, G, B channel values into separate arrays. */
	for (i = 0; i < h; i++) {
		for (j = 0; j < w; j++) {
			r[i*w+j] = pp[0];
			g[i*w+j] = pp[1];
			b[i*w+j] = pp[2];
			pp += rowpix;
		}
	}

	/* Run filter on RGB channels. */
	filter_image(r, rout, w, h, gauss, 5);
	filter_image(g, gout, w, h, gauss, 5);
	filter_image(b, bout, w, h, gauss, 5);

	pp = pix;

	/* Write filtered values back to buffer. */
	for (i = 0; i < h; i++) {
		for (j = 0; j < w; j++) {
			pp[0] = rout[i*w+j];
			pp[1] = gout[i*w+j];
			pp[2] = bout[i*w+j];
			pp += rowpix;
		}
	}

	/* Save image buffer to PNG file. */
	gerror = NULL;
	ok = gdk_pixbuf_save(pbuf, "out.png", "png", &gerror, NULL);

	if (!ok)
		printf("can't save file: %s\n", gerror->message);

	/* This is not necessary since we exit now... */
	free(rout);
	free(r);

	g_object_unref(pbuf);

	return 0;
}

